from mxcubecore import HardwareRepository as HWR

from mxcubeweb.core.components.component_base import ComponentBase


class Beamline(ComponentBase):
    def __init__(self, app, config):
        super().__init__(app, config)

    def init_signals(self):
        pass

    def get_viewport_info(self):
        """
        Get information about current "view port" video dimension, beam position,
        pixels per mm, returns a dictionary with the format:

            data = {"pixelsPerMm": pixelsPerMm,
                    "imageWidth": width,
                    "imageHeight": height,
                    "format": fmt,
                    "sourceIsScalable": source_is_scalable,
                    "scale": scale,
                    "videoSizes": video_sizes,
                    "position": position,
                    "shape": shape,
                    "size_x": sx, "size_y": sy}

        :returns: Dictionary with view port data, with format described above
        :rtype: dict
        """
        fmt, source_is_scalable = "MJPEG", False

        if self.app.CONFIG.app.VIDEO_FORMAT == "MPEG1":
            fmt, source_is_scalable = "MPEG1", True
            video_sizes = HWR.beamline.sample_view.camera.get_available_stream_sizes()
            (width, height, scale) = HWR.beamline.sample_view.camera.get_stream_size()
        else:
            scale = 1
            width = HWR.beamline.sample_view.camera.get_width()
            height = HWR.beamline.sample_view.camera.get_height()
            video_sizes = [(width, height)]

        pixelsPerMm = HWR.beamline.diffractometer.get_pixels_per_mm()

        return {
            "pixelsPerMm": pixelsPerMm,
            "imageWidth": width,
            "imageHeight": height,
            "format": fmt,
            "sourceIsScalable": source_is_scalable,
            "scale": scale,
            "videoSizes": video_sizes,
            "videoHash": HWR.beamline.sample_view.camera.stream_hash,
            "videoURL": self.app.CONFIG.app.VIDEO_STREAM_URL,
        }



import logging
from typing import ClassVar

import pydantic
from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.Beamline import Beamline

from mxcubeweb.core.adapter.adapter_base import ActuatorAdapterBase
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel

resource_handler_config = ResourceHandlerConfigModel(
    commands=["prepare_beamline_for_sample"],
    attributes=["data", "get_value"],
)


class BeamlineAdapter(ActuatorAdapterBase):
    """
    Adapter between Beamline route and Beamline hardware object.
    """

    SUPPORTED_TYPES: ClassVar[list[object]] = [Beamline]

    def __init__(self, ho, role, app):
        super().__init__(ho, role, app, resource_handler_config)
        self.app = app
        self.adapter_dict = {}

        workflow = self._ho.workflow
        if workflow:
            workflow.connect("parametersNeeded", self._wf_parameters_needed)

        gphl_workflow = self._ho.gphl_workflow
        if gphl_workflow:
            gphl_workflow.connect(
                "GphlJsonParametersNeeded", self._gphl_json_wf_parameters_needed
            )
            gphl_workflow.connect(
                "GphlUpdateUiParameters", self._gphl_json_wf_update_ui_parameters
            )

        from mxcubeweb.routes import signals

        if HWR.beamline.xrf_spectrum:
            HWR.beamline.xrf_spectrum.connect(
                HWR.beamline.xrf_spectrum,
                "xrf_task_progress",
                signals.xrf_task_progress,
            )

    def _wf_parameters_needed(self, params):
        self.app.server.emit("workflowParametersDialog", params, namespace="/hwr")

    def _gphl_json_wf_parameters_needed(self, schema, ui_schema):
        params = {}
        params["schema"] = schema
        params["ui_schema"] = ui_schema
        self.app.server.emit("gphlWorkflowParametersDialog", params, namespace="/hwr")

    def _gphl_json_wf_update_ui_parameters(self, update_dict):
        self.app.server.emit(
            "gphlWorkflowUpdateUiParametersDialog", update_dict, namespace="/hwr"
        )

    def _get_available_elements(self):
        escan = self._ho.energy_scan
        return escan.get_elements() if escan else []

    def get_value(self) -> dict:
        """
        Build dictionary value-representation for each beamline attribute
         Returns:
           (dict): The dictionary.
        """
        attributes = {}

        for attr_name in self.app.mxcubecore.adapter_dict:
            # We skip the beamline attribute to avoid endless recursion
            if attr_name == "beamline":
                continue

            try:
                _d = self.app.mxcubecore.get_adapter(attr_name).data().dict()
            except pydantic.ValidationError:
                logging.getLogger("MX3.HWR").error(f"Incorrect values in {attr_name}")
                logging.getLogger("MX3.HWR").exception("")

            attributes.update({attr_name: _d})

        return {
            "energyScanElements": self._get_available_elements(),
            "path": HWR.beamline.session.get_base_image_directory(),
            "actionsList": [],
            "hardwareObjects": attributes,
        }

    def prepare_beamline_for_sample(self) -> dict:
        if hasattr(HWR.beamline.collect, "prepare_for_new_sample"):
            HWR.beamline.collect.prepare_for_new_sample()

        return {}



import io
import json
import logging

from flask import (
    Blueprint,
    Response,
    jsonify,
    request,
    send_file,
)
from mxcubecore import HardwareRepository as HWR


# Disabling C901 function is too complex (19)
def init_route(app, server, url_prefix):  # noqa: C901
    bp = Blueprint("sampleview", __name__, url_prefix=url_prefix)

    @bp.route("/camera/subscribe", methods=["GET"])
    @server.restrict
    def subscribe_to_camera():
        """
        Subscribe to the camera streaming
            :response: image as html Content-type
        """
        if app.CONFIG.app.VIDEO_FORMAT == "MPEG1":
            result = Response(status=200)
        else:
            result = app.sample_view.http_streamer.get_response()

        return result

    @bp.route("/camera/unsubscribe", methods=["PUT"])
    @server.restrict
    def unsubscribe_to_camera():
        """
        SampleCentring: unsubscribe from the camera streaming
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        HWR.beamline.sample_view.camera.streaming_greenlet.kill()
        return Response(status=200)

    @bp.route("/camera/snapshot", methods=["POST"])
    @server.restrict
    def snapshot():
        """
        Take snapshot of the sample view
        data = {"overlay": overlay_data} overlay is the image data to overlay on sample image,
        it should normally contain the data of shapes drawn on canvas.
        Return: Overlayed image uri, if successful, statuscode 500 otherwise.
        """
        try:
            overlay = json.loads(request.data).get("overlay")
            mimetype, overlay_data = overlay.split(",")

            # Check if send data is a jpeg image
            if "image/jpeg" not in mimetype:
                msg = "Image type should be jpeg"
                raise ValueError(msg)

            image = HWR.beamline.sample_view.take_snapshot(
                overlay_data=overlay_data,
            )

            b = io.BytesIO()
            image.save(b, "JPEG")
            b.seek(0)

            return send_file(
                b,
                mimetype="image/jpeg",
                as_attachment=True,
                download_name="snapshot.jpeg",
            )
        except Exception:
            logging.getLogger("MX3.HWR").exception("Taking a snapshot failed")
            return jsonify({"error": "Taking a snapshot failed"})

    @bp.route("/camera", methods=["GET"])
    @server.restrict
    def get_image_data():
        """
        Get size of the image of the diffractometer
            :response Content-type:application/json, example:
            {  "imageHeight": 576, "imageWidth": 768,
            "pixelsPerMm": [1661.1295681063123, 1661.1295681063123]
            }
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        data = app.beamline.get_viewport_info()

        resp = jsonify(data)
        resp.status_code = 200
        return resp

    @bp.route("/camera", methods=["POST"])
    @server.restrict
    def set_image_size():
        """ """
        params = request.get_json()

        res = app.sample_view.set_image_size(
            float(params["width"]), float(params["height"])
        )

        resp = jsonify(res)
        resp.status_code = 200
        return resp

    @bp.route("/centring/<point_id>/moveto", methods=["PUT"])
    @server.require_control
    @server.restrict
    def move_to_centred_position(point_id):
        """
        Move to the given centred position.
            :parameter id: centred position identifier, integer
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        point = app.sample_view.move_to_centred_position(point_id)

        if point:
            return Response(status=200)
        return Response(status=409)

    @bp.route("/shapes", methods=["GET"])
    @server.restrict
    def get_shapes():
        """
        Retrieve all the stored centred positions.
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        shapes = app.sample_view.get_shapes()

        resp = jsonify(shapes)
        resp.status_code = 200
        return resp

    @bp.route("/shapes", methods=["POST"])
    @server.require_control
    @server.restrict
    def update_shapes():
        """
        Update shape information.
            :parameter shape_data: dict with shape information (id, type, ...)
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        shapes = request.get_json().get("shapes", [])

        resp = jsonify(app.sample_view.update_shapes(shapes))
        resp.status_code = 200

        return resp

    @bp.route("/shapes/<sid>", methods=["DELETE"])
    @server.require_control
    @server.restrict
    def delete_shape(sid):
        """
        Retrieve all the stored centred positions.
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        HWR.beamline.sample_view.delete_shape(sid)
        return Response(status=200)

    @bp.route("/shapes/rotate_to", methods=["POST"])
    @server.require_control
    @server.restrict
    def rotate_to():
        """
        Rotate Phi to the position where the given shape was defined

            :parameter sid: The shape id
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        sid = request.get_json().get("sid", -1)

        try:
            app.sample_view.rotate_to(sid)
        except Exception:
            resp = Response(status=409)
        else:
            resp = Response(status=200)

        return resp

    @bp.route("/centring/start_click_centring", methods=["PUT"])
    @server.require_control
    @server.restrict
    def centre_click():
        """
        Start Click centring procedure.
            :statuscode: 200: no error
            :statuscode: 409: error
        """

        try:
            data = app.sample_view.start_manual_centring()
        except Exception:
            msg = "Could not start %s click centring"
            logging.getLogger("MX3.HWR").exception(
                msg, HWR.beamline.config.click_centring_num_clicks
            )
            resp = (
                "Could not move motor",
                409,
                {"Content-Type": "application/json"},
            )
        else:
            resp = jsonify(data)
            resp.status_code = 200

        return resp

    @bp.route("/centring/abort", methods=["PUT"])
    @server.require_control
    @server.restrict
    def abort_centring():
        """
        Abort centring procedure.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        app.sample_view.abort_centring()
        return Response(status=200)

    @bp.route("/centring/click", methods=["PUT"])
    @server.require_control
    @server.restrict
    def click():
        """
        The click method needs the input from the user, a running click centring
        procedure must be set before

        :request Content-type: application/json, integer positions of the clicks,
                            {clickPos={"x": 123,"y": 456}}
        :response Content-type: application/json, integer, number of clicks
                                left {'clickLeft': 3 | 2 | 1}
        :statuscode: 200: no error
        :statuscode: 409: error
        """
        pos = json.loads(request.data).get("clickPos", None)

        data = app.sample_view.centring_handle_click(pos["x"], pos["y"])

        resp = jsonify(data)
        resp.status_code = 200
        return resp

    @bp.route("/centring/accept", methods=["PUT"])
    @server.require_control
    @server.restrict
    def accept_centring():
        """
        Accept the centring position.
        """
        HWR.beamline.diffractometer.accept_centring()
        return Response(status=200)

    @bp.route("/movetobeam", methods=["PUT"])
    @server.require_control
    @server.restrict
    def move_to_beam():
        """Go to the beam position from the given (x, y) position."""
        pos = json.loads(request.data).get("clickPos")

        app.sample_view.move_to_beam(pos["x"], pos["y"])

        return Response(status=200)

    @bp.route("/centring/centring_method", methods=["PUT"])
    @server.require_control
    @server.restrict
    def set_centring_method():
        """
        Set MXCuBE to use automatic (lucid) centring procedure when
        mounting samples

        :statuscode: 200: no error
        :statuscode: 409: error

        """
        method = json.loads(request.data).get("centringMethod", None)
        app.sample_view.set_centring_method(method)
        return Response(status=200)

    return bp



import logging
from enum import Enum
from typing import ClassVar

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.BeamlineActions import BeamlineActions
from pydantic.v1 import (
    BaseModel,
)

from mxcubeweb.core.adapter.adapter_base import AdapterBase
from mxcubeweb.core.components.queue import (
    FAILED,
    READY,
    RUNNING,
)
from mxcubeweb.core.models.adaptermodels import (
    BeamlineActionInputModel,
    HOActuatorValueChangeModel,
    NStateModel,
)
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel

resource_handler_config = ResourceHandlerConfigModel(
    commands=["stop", "run_action"], attributes=["data", "get_all_actions"]
)


class Argument(BaseModel):
    name: str
    type: str
    items: list[str] | None = None


class JSONSchemaArgument(BaseModel):
    name: str
    type: str
    jsonschema: str = ""
    signature: list[str]


class Action(BaseModel):
    name: str
    username: str
    state: int
    arguments: list[JSONSchemaArgument | Argument]
    argument_type: str
    messages: list[str]
    type: str
    data: str | dict | None


class ActionsList(BaseModel):
    actions_list: list[Action]


class BeamlineActionAdapter(AdapterBase):
    SUPPORTED_TYPES: ClassVar[list[object]] = [BeamlineActions]

    def __init__(
        self,
        ho: HardwareObject,
        role: str,
        app,
        resource_handler_config: ResourceHandlerConfigModel = resource_handler_config,
    ):
        """
        Args:
            (object): Hardware object.
        """
        super().__init__(ho, role, app, resource_handler_config)
        self._value_change_model = HOActuatorValueChangeModel

        ho.connect("valueChanged", self._value_change)
        ho.connect("stateChanged", self.state_change)

        for cmd in ho.get_commands() + ho.get_annotated_commands():
            cmd.connect(
                "commandBeginWaitReply",
                self._beamline_action_start,
            )
            cmd.connect(
                "commandReplyArrived",
                self._beamline_action_done,
            )
            cmd.connect(
                "commandFailed",
                self._beamline_action_failed,
            )

    def _value_change(self, value):
        v = value.name if isinstance(value, Enum) else value
        self.value_change(v)

    def _emit_beamline_action(self, msg):
        self.app.server.emit("beamline_action", msg, namespace="/hwr")

    def _beamline_action_start(self, name):
        msg = {"name": name, "state": RUNNING}
        self._emit_beamline_action(msg)
        logging.getLogger("user_level_log").info("Command %s started.", name)

    def _beamline_action_done(self, name, result):
        msg = {"name": name, "state": READY, "data": result}
        self._emit_beamline_action(msg)
        logging.getLogger("user_level_log").info("Command %s done.", name)

    def _beamline_action_failed(self, name):
        msg = {"name": name, "state": FAILED}
        self._emit_beamline_action(msg)
        logging.getLogger("user_level_log").error("Action %s failed!", name)

    def msg(self):
        return ""

    def stop(self):
        """
        Stop the execution.
        """
        for cmd in self._ho.get_commands():
            self._ho.abort_command(cmd.name())

    def _valid_action_input(self, value: dict | list) -> bool:
        """
        Validates that the the action input value structure is
        flat and contains only valid types (str, int, float, bool).

        Args:
            value The structure to validate.
        Returns:
            bool: True if valid, False otherwise.
        """
        allowed_types = (str, int, float, bool)

        if isinstance(value, list):
            return all(isinstance(item, allowed_types) for item in value)
        if isinstance(value, dict):
            return all(isinstance(item, allowed_types) for item in value.values())
        return False

    def run_action(self, value: BeamlineActionInputModel):
        """
        Starts beamline action
        """
        # Beamline actions are retrieved from a finite list of commands
        # its either a simple command name or an annotated command
        # Getting the KeyError means that the command is not annotated
        try:
            self._ho.get_annotated_command(value.cmd)
        except KeyError:
            annotated_cmd = False
        else:
            annotated_cmd = True

        # Annotated commands are validated against their Pydantic model
        # Simple comamands are validated with _valid_action_input
        if not annotated_cmd and self._valid_action_input(value.parameters) is False:
            msg = (
                f"Action '{value.cmd}' cannot run: parameters must contain"
                " (str, int, float, bool)"
                f" but got '{type(value.parameters)}'"
            )
            logging.getLogger("MX3.HWR").error(msg)
            raise ValueError(msg)

        try:
            self._ho.execute_command(value.cmd, value.parameters)
        except Exception:
            msg = f"Action cannot run: command {value.cmd} does not exist"
            logging.getLogger("MX3.HWR").exception(msg)
            raise

    def get_all_actions(self) -> ActionsList:
        actions: list[Action] = []

        try:
            cmds = self._ho.get_commands()
        except Exception:
            cmds = []

        for cmd in cmds:
            args: list[Argument] = []
            for arg in cmd.get_arguments():
                argname = arg[0]
                argtype = arg[1]

                argument_data = {"name": argname, "type": argtype}
                if argtype == "combo":
                    argument_data["items"] = cmd.get_combo_argument_items(argname)
                args.append(Argument(**argument_data))

            action = Action(
                name=cmd.name(),
                username=cmd.name(),
                state=READY,
                arguments=args,
                argument_type=cmd.argument_type,
                messages=[],
                type=cmd.type,
                data=cmd.value(),
            )
            actions.append(action)

        if getattr(self._ho, "pydantic_model", None):
            for cmd_name in self._ho.exported_attributes:
                cmd_object = self._ho.get_annotated_command(cmd_name)
                exported = self._ho.exported_attributes[cmd_name]

                json_arg = JSONSchemaArgument(
                    name=cmd_name,
                    type="JSONSchema",
                    jsonschema=exported["schema"],
                    signature=exported["signature"],
                )

                action = Action(
                    name=cmd_name,
                    username=cmd_object.name(),
                    state=READY,
                    arguments=[json_arg],
                    argument_type="JSONSchema",
                    jsonschema=exported["schema"],
                    messages=[],
                    type="JSONSchema",
                    data="",
                )
                actions.append(action)

        return ActionsList(actions_list=actions)

    def data(self) -> NStateModel:
        return NStateModel(**self._dict_repr())
